/* i5I.c */
/* programs for inputting and printing MPI's. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "integer.h"
#include "fun.h"

/*
This code implements a memory bank for allocated (but finished with) MPI's.
Please see the code below for more comments on what is happening.
		Peter Adams, 1/4/1993.
*/

/*
Only maintain 4 MPIBANKS with at most 500 MPI's in each bank on a PC, due to
memory restrictions.
Maintain 20 MPIBANKS with no size limit on other machines
*/
#ifdef _WIN32
#define NMPIBANKS 4
#define MAXMPIBSIZE 500
int MPI_BANKSIZE[NMPIBANKS];
#else
#define NMPIBANKS 20
#endif

int ctr[100];
MPI *MPI_BANK[NMPIBANKS];

/*
The MPI Bank works like this:
MPI's are of different sizes. Hence it does not work to place all free'd MPI's
into the one bank. Now (in theory) MPI's can be of ANY size. However, in most
cases they are small. Hence much malloc'ing and free'ing was being done on
MPI's of the same small sizes. 

We maintain n=NMPIBANKS of free'd MPI's. Each of these banks is just a linked
list of MPI's of that size. 
*/

void INITMPIBANK()
/*
Initially, set the banks to being empty (a null pointer). On a PC, because we
restrict the length allowed for each list, set a counter to 0.
*/
{
	int i;

	for(i = 0; i < NMPIBANKS; i++)
	{
	  MPI_BANK[i] = NULL;
#ifdef _WIN32
	  MPI_BANKSIZE[i] = 0;
#endif
	}
}

MPI *BUILDMPI(unsigned int n)
/*
 * mallocs space for an MPI of length n.
 * If there is an MPI in the bank of this size, then use it rather than malloc.
 */
{
	MPI *Mptr;
	unsigned int m;

	m = n - 1;
	if(m < NMPIBANKS && MPI_BANK[m])
	{
	  Mptr = MPI_BANK[m];
	  MPI_BANK[m] = MPI_BANK[m]->NEXT;
#ifdef _WIN32
	  MPI_BANKSIZE[m]--;
#endif
	}
	else
	{
	  Mptr = (MPI *)mmalloc(sizeof(MPI));
	  Mptr->V = (unsigned int *)mmalloc(n * sizeof(unsigned int));
	  Mptr->D = n - 1;
	}
	return (Mptr);
}

void FREEMPIBANK()
/*
This routine should be called to clear out the MPI bank, usually at the end of
execution. If you like, you can call it during execution, and all of the MPI's
in the list will be free'd up. 
The only time you will probably want to call this during execution is if the
algorithm just performed a lot of very memory intensive operations, and so there
will be many many MPI's on the list. Freeing them up will free up some memory,
but it will mean that subsequent calls to create MPI's will need to perform the
(slow) malloc command until there are some more free ones.
*/
{
	int i;
	MPI *tmp,*this;
	unsigned int m;

	for(i=0; i<NMPIBANKS; i++)
	{
	  this=MPI_BANK[i];
	  while(this)
	  {
	    tmp=this->NEXT;
	    m = this->D;
	    ffree ((char *)(this->V), (1 + m) * sizeof(unsigned int));
	    ffree ((char *)this, sizeof(MPI));
	    this=tmp;
	  }
	  MPI_BANK[i]=NULL;
	}
	return;
}

void FREEMPI(MPI *Mptr)
/*
 * deallocates the space alloted for the MPI Mptr.
 * This involves placing it onto the appropriate MPI Bank, if there is such a
 * bank, and (on a PC) if there is room in the bank.
 * If it doesn't go into a bank, then it is free'd.
 */
{
	unsigned int m;

	m = Mptr->D;

	if(m < NMPIBANKS) 
	{
#ifdef _WIN32
	  if(MPI_BANKSIZE[m] == MAXMPIBSIZE)
	  {
	    ffree (Mptr->V, (1 + m) * sizeof(unsigned int));
	    ffree (Mptr, sizeof(MPI));
	  }
	  else
	  {
	    Mptr->NEXT = MPI_BANK[m];
	    MPI_BANK[m] = Mptr;
	    MPI_BANKSIZE[m]++;
	  }
#else
	  Mptr->NEXT = MPI_BANK[m];
	  MPI_BANK[m] = Mptr;
#endif
	}
	else
	{
	  ffree (Mptr->V, (1 + m) * sizeof(unsigned int));
	  ffree (Mptr, sizeof(MPI));
	}
}

MPI *COPYI(MPI *Aptr)
/*
 * returns a pointer to a copy of *Aptr.
 */
{
	MPI *Bptr;

	Bptr = BUILDMPI(1 + Aptr->D);
	Bptr->S = Aptr->S;
	memcpy(Bptr->V, Aptr->V, (1 + Aptr->D) * sizeof(unsigned int));
	return Bptr;
}

void FPRINTI(FILE *outfile, MPI *Mptr)
/*
 * the MPI *Mptr is printed in decimal notation to *outfile.
 * no new-line is incorporated. 
 */
{
	MPI *F, *Temp;
	unsigned int  *A, l;
	int i = 0;
	
	if (Mptr->S == 0) 
	{
		fprintf(outfile, "%u", 0);
		return;
	}
	if (Mptr->S == -1) 
		fprintf(outfile, "-");
	F = ABSI(Mptr);
	l = LENGTHI(F);
	A = (unsigned int *)mmalloc(l * sizeof(unsigned int));
	do 
	{
		A[i] = MOD0_(F, 10);
		Temp = F;
		F = INT0_(F, 10);
		FREEMPI(Temp);
		i++;
	}
        while (F->S);
	i--;
	FREEMPI(F);
	while (i >= 0) 
	{
		fprintf(outfile, "%u", A[i]);
		i--;
	}
	ffree((char *)A, l * sizeof(unsigned int)); 
	return;
}

void PRINTI(MPI *Mptr)
{
	FPRINTI(stdout, Mptr);
	return;
}

MPI *ZEROI()
/*
 * Returns the MPI ZERO.
 */
{
	MPI *Eptr;

	Eptr = BUILDMPI(1);
	Eptr->S = 0;
	Eptr->V[0] = 0;
	return Eptr;
}

MPI *ONEI()
/*
 * Returns the MPI ONE.
 */
{
	MPI *Eptr;

	Eptr = BUILDMPI(1);
	Eptr->S = 1;
	Eptr->V[0] = 1;
	return Eptr;
}

MPI *MINUS_ONEI()
/*
 * Returns the MPI MINUS_ONE.
 */
{
	MPI *Eptr;

	Eptr = BUILDMPI(1);
	Eptr->S = -1;
	Eptr->V[0] = 1;
	return Eptr;
}

MPI *CHANGE(USI n)
/*
 * converts n, 0 <= n < R0 to an MPI *Mptr. 
 */
{
	MPI *Mptr;

	Mptr = BUILDMPI(1);
	if (n == 0)
        {
		Mptr->S = 0;
		Mptr->V[0] = 0;
		return Mptr;
	}
	else 
	{
		Mptr->S = 1;
		Mptr->V[0] = n;
		return Mptr;
	}
}

MPI *CHANGEI(int n)
/*
 * converts n, 0 <= |n| < R0 to an MPI *Mptr. 
 */
{
	MPI *Mptr;
	unsigned int m;

	Mptr = BUILDMPI(1);
	m = abs(n);
	Mptr->V[0] = m;
	if (n == 0)
		Mptr->S = 0;
	if (n < 0)
		Mptr->S = -1;
	else 
		Mptr->S = 1;
	return Mptr;
}

MPI *FINPUTI(FILE *f, USI *uptr)
/*
 * Converts decimal input from stream into an MPI.
 * Ignores the combination '\' followed by '\n'.
 * If a rubbish character is met before decimal input, Mptr is set to 0
 * and 0 is returned. All characters up to and including the first newline
 * met are wiped.
 * If a rubbish character is met immediately after decimal input, 0 is
 * returned and all characters up to and including the first newline met
 * are wiped. Otherwise 1 is returned.
 * In any case *Mptr is set equal to any inputted decimal.
 */
{
	unsigned int n;
	int s, t;
	MPI *Mptr, *Temp;

	Mptr = BUILDMPI(1);
	Mptr->S = 0;
	Mptr->V[0] = 0;
	s = whitespace(f);
	if ((s != '-') && (s < '0' || s > '9')) 
	{
		printf("illegal character %c entered:\n", s);
		FFlush(f);
		*uptr = 0;
		return Mptr;
	}
	else if (s == '0')
	{
		s = fgetc(f);
		if (s != ' ' && s != '\n' && s != 'i' && s != '/' && s != '+' && s != '-')
		{
			printf("illegal character %c entered:\n", s);
			FFlush(f);
			*uptr = 0;
			return Mptr;
		}
		ungetc(s, f);
		*uptr = 1;
		return Mptr;
	}
	else if (s == '-') 
	{
		do
			s = fgetc(f);
		while (s == ' ');
		if (s <= '0' || s > '9')
		{
			printf("illegal character %c entered:\n", s);
			FFlush(f);
			*uptr = 0;
			return Mptr;
		}
		t = -1;
	}
        else
		t = 1;
	while ((s == '\\') || (s >= '0' && s <= '9')) 
	{
		if (s == '\\')
		/* for use with bc output, where '\' is followed by '\n' */
			s = fgetc(f);
		else 
		{
			Temp = Mptr;
			Mptr = MULT_I(Temp, 10);
			FREEMPI(Temp);
			n = (unsigned int)(s - '0');
			Temp = Mptr;
			Mptr = ADD0_I(Temp, n);
			FREEMPI(Temp);
		}
	 	s = fgetc(f);
	}
	if (t == -1)
		Mptr->S = -1;
	if (s != ' ' && s != '\n' && s != 'i' && s != '/' && s != '+' && s != '-')
	{
		printf("illegal character %c entered:\n", s);
		FFlush(f);
		*uptr = 0;
		return Mptr;
	}
	ungetc(s, f);
	*uptr = 1;
	return Mptr;
}

MPI *INPUTI(USI *uptr)
{
	return  FINPUTI(stdin, uptr);
}

unsigned int EQUALI (MPI *Aptr, MPI *Bptr)
/*
 * EQUALI(Aptr, Bptr) = 1 if *Aptr = *Bptr, otherwise  = 0.
 */
{
	int j;

	if (Aptr->S != Bptr->S)
		return (0);
	else if (Aptr->D != Bptr->D)
		return (0);
	else
	{
		j = Aptr->D;
		while ((Aptr->V[j] == Bptr->V[j]) && (j > 0))
			j--;
		if (Aptr->V[j] == Bptr->V[j])
			return (1);
		else
			return (0);
	}
}

MPI *ABSI(MPI *Aptr)
/*
 * *Bptr = |*Aptr|. 
 */
{
	MPI *Bptr;

	Bptr = COPYI(Aptr);
	Bptr->S = (Aptr->S >= 0 ? Aptr->S : -Aptr->S);
	return Bptr;
}

MPI *MINUSI(MPI *Aptr)
/*
 * *Bptr = -(*Aptr).
 */
{
	MPI *Bptr;

	Bptr = COPYI(Aptr);
	Bptr->S = -(Aptr->S);
	return Bptr;
}

MPI *GCD(MPI *Aptr, MPI *Bptr)
/*
 * GCD(|*Aptr|, |*Bptr|). 
 */
{
	MPI *A, *B, *Cptr, *Temp;

	A = ABSI(Aptr);
	if (Bptr->S == 0)
		return A;
	B = ABSI(Bptr);
	Cptr = MOD0(A, B);
	while (Cptr->S == 1)
	{
		Temp = A;
		A = B;
		FREEMPI(Temp);
		B = Cptr;
		Cptr = MOD0(A, B);
	}
	Temp = Cptr;
	Cptr = B;
	FREEMPI(Temp);
	FREEMPI(A);
	return Cptr;
}

MPI *LCM(MPI *Aptr, MPI *Bptr)
/*
 * *Cptr = LCM(*Aptr, *Bptr). 
 */
{
	MPI *C, *N, *Cptr, *Temp;

	if (Aptr->S == 0 || Bptr->S == 0)
		return ZEROI();
	else
	{
		N = GCD(Aptr, Bptr);
		C = MULTI(Aptr, Bptr);
		Temp = C;
		C = ABSI(Temp);
		FREEMPI(Temp);
		Cptr = INT0(C, N);
		FREEMPI(C);
		FREEMPI(N);
		return Cptr;
	}
}

MPI *LCM_ARRAY(MPI *M[], USI n)
/*
 * *Aptr = LCM(M[0], ..., M[n - 1]).
 */
{
	MPI *L, *Aptr;
	unsigned int i;

	Aptr = COPYI(M[0]);
	if (n == 1)
		return Aptr;
	else
	{
		for ( i = 1; i <= n - 1; i++)
		{
			L = Aptr;
			Aptr = LCM(L, M[i]);
			FREEMPI(L);
		}
		return Aptr;
	}
}

MPI *LCM_MATRIX(MPMATI *Mptr)
/*
 * *Aptr = LCM(Mptr->V[i][j]), 0 <= i <= Mptr->R - 1, 0 <= j <= Mptr->C - 1.
 */
{
	MPI *Aptr;
	unsigned int m, n;

	m = Mptr->R;
	n = Mptr->C;
	Aptr = LCM_ARRAY(Mptr->V, m * n);
		return Aptr;
}

void CONT_FRAC(MPI *Aptr, MPI *Bptr, MPI *Zptr, MPI **Xptr, MPI **Yptr)
/*
 * Euclid's algorithm for use in 2SQUARES. 
 */
{
	MPI *A, *B, *C, *Temp;

	A = COPYI(Aptr);
	B = COPYI(Bptr);
	C = MOD0(A, B);
	while (C->S)
	{
		if (RSV(C, Zptr) <= 0)
		{
			*Xptr = COPYI(C);
			Temp = A;
			A = B;
			FREEMPI(Temp);
			B = C;
			*Yptr = MOD0(A, B);
			FREEMPI(A);
			FREEMPI(B);
			return;
		}
		Temp = A;
		A = B;
		FREEMPI(Temp);
		B = C;
		C = MOD0(A, B);
	}
}

void INTSET(USI *Aptr, USI n, USI m)
/*
 * this converts the n array elements *Aptr,....*(Aptr + n - 1) to m. 
 */
{
	for( ; n; n--)
		*(Aptr++) = m;
	return;
}

void INTSETI(int *Aptr, USI n, int m)
/*
 * this converts the n array elements *Aptr,....*(Aptr + n - 1) to m. 
 */
{
	for( ; n; n--)
		*(Aptr++) = m;
	return;
}

unsigned int LENGTHI(MPI *Mptr)
/*
 * returns the number of decimal digits in the MPI *Mptr,
 * increased by 1 if *Mptr is negative.
 */
{
	MPI *F, *Temp;
	unsigned int i = 0, s;
	
	if (Mptr->S >= 0)
		s = 0;
	else
		s = 1;
	F = ABSI(Mptr);
	do
	{
		Temp = F;
		F = INT0_(Temp, 10);
		FREEMPI(Temp);
		i++;
	} while (F->S == 1);
	FREEMPI(F);
	return (i + s);
}

unsigned int CONVERTI(MPI *N)
/*
 * Returns N as an unsigned integer, providing N is small enough.
 * If N's too big, you get an overflow error.
 */
{
	if (N->D > 1)
	{
		fprintf(stderr, "overflow in CONVERTI\n");
		exit(1);
	}
	if (N->S == 0)
		return 0;
	if (N->D == 0)
		return N->V[0];
	else
		return ((N->V[1]) * R0 + (N->V[0]));
}

MPI *INPUTSI(char **ptr, unsigned int *uptr)
/*
 * Converts decimal input from a stream into an MPI, leaving *ptr to point
 * to the character following the last decimal digit.
 * Ignores the combination '\' followed by '\n'.
 * If a rubbish character is met before decimal input, Mptr is set to 0
 * and *uptr = 0 is returned.
 * If a rubbish character is met immediately after decimal input, 0 is
 * returned,  Otherwise *uptr = 1 is returned.
 * In any case *Mptr is set equal to any inputted decimal.
 */
{
	unsigned int n;
	int t;
	MPI *Mptr, *Temp;

	Mptr = ZEROI();
	while (**ptr == ' ' || **ptr == '\n')
		(*ptr)++;
	if (**ptr != '-' && (**ptr < '0' || **ptr > '9')) 
	{
		if (**ptr != '.')
		{
		printf("in INPUTSI1, illegal character %c entered:\n", **ptr);
		}
		*uptr = 0;
		return (Mptr);
	}
	else if (**ptr == '0')
	{
		(*ptr)++;
		if (**ptr != ' ' && **ptr != '\n' && **ptr != 'i' && **ptr != '/' && **ptr != '+' && **ptr != '-' && **ptr != '\0')
		{
			if (**ptr != '.')
			{
			printf("in INPUTSI2, illegal character %c entered:\n", **ptr);
			}
			*uptr = 0;
			return (Mptr);
		}
		*uptr = 1;
		return (Mptr);
	}
	else if (**ptr == '-') 
	{
		do
			(*ptr)++;
		while (**ptr == ' ');
		if (**ptr <= '0' || **ptr > '9')
		{
			printf("in INPUTSI3 illegal character %c entered:\n", **ptr);
			*uptr = 0;
			return (Mptr);
		}
		t = -1;
		while (**ptr == ' ')
			(*ptr)++;
	}
        else
		t = 1;
	while ((**ptr == '\\') || (**ptr >= '0' && **ptr <= '9')) 
	{
		if (**ptr == '\\')
		/* for use with bc output, where '\' is followed by '\n'*/
			(*ptr)++;
		else 
		{
			Temp = Mptr;
			Mptr = MULT_I(Temp, 10);
			FREEMPI(Temp);
			n = (unsigned int)(**ptr - '0');
			Temp = Mptr;
			Mptr = ADD0_I(Temp, n);
			FREEMPI(Temp);
		}
		(*ptr)++;
	}
	if (t == -1)
		Mptr->S = -1;
	if (**ptr != ' ' && **ptr != '\n' && **ptr != 'i' && **ptr != '/' && **ptr != '+' && **ptr != '-' && **ptr != '\0')
	{
		if (**ptr != '.')
		{
		printf("in INPUTSI4 illegal character %c entered:\n", **ptr);
		}
		*uptr = 0;
		return (Mptr);
	}
	*uptr = 1;
	return (Mptr);
}

void SPRINTI(char *buffer, MPI *Mptr)
/*
 * The decimal digits of the MPI *Mptr are placed in the string buffer.
 * no new-line is incorporated. 
 */
{
	MPI *F, *Temp;
	unsigned int  *A,l ;
	int i = 0;
	
	if (Mptr->S == 0) 
	{
		sprintf(buffer, "%u", 0);
		return;
	}
	if (Mptr->S == -1) 
	{
		sprintf(buffer, "-");
		buffer++;
	}
	F = ABSI(Mptr);
	l = LENGTHI(F);
	A = (unsigned *)mmalloc(l * sizeof(unsigned int));
	do
        {
		A[i] = MOD0_(F, 10);
		Temp = F;
		F = INT0_(Temp, 10);
		FREEMPI(Temp);
		i++;
	}
 	while (F->S);
	i--;
	while (i >= 0)
        {
		sprintf(buffer, "%u", A[i]);
		buffer++;
		i--;
	}
	ffree((char *)A, l * sizeof(unsigned int)); 
	FREEMPI(F);
	return;
}

MPMATI *FINPUTMATI(FILE *infile)
/*
 * Inputs a matrix of MPI's from infile.
 */
{
	MPMATI *Temp;
	unsigned int u, i, j, m, n;
	char *line, *p;

	if (infile == stdin)
		fprintf(stderr, "Enter the number of rows and number of columns\n");

	if (fscanf(infile, "%u%u", &m, &n) != 2 || m == 0 || n == 0)
	{
		fprintf(stderr, "corrupt input: returning the 1 x 1 zero matrix\n");
		if (infile != stdin)
			exit (1);
		return ZEROMNI(1, 1);
	}
	fgetc(infile);
	Temp = ZEROMNI(m, n);
	for (i = 0; i < m; i++)
	{
		if (infile == stdin)
			fprintf(stderr, "Enter row %u:\n", i + 1);
		p = line = Fgets(infile);
		for (j = 0; j < n; j++)
		{
			FREEMPI(elt(Temp, i, j));
			elt(Temp, i, j) = INPUTSI(&p, &u);
			if (u == 0)
			{
				free(line);
				fprintf(stderr, "matrix input interrupted prematurely\n");
				if (infile != stdin)
					exit (1);
				GetReturn();
				return Temp;
			}
		}
		free(line);
	}
	return (Temp);
}

MPMATI *INPUTMATI()
/*
 * Input a matrix of MPI's from stdin.
 */
{
	return FINPUTMATI(stdin);
}

void PRINTMATI(USI i1, USI i2, USI j1, USI j2, MPMATI *Mptr)
/*
 * prints *Mptr from rows i1 to i2 and cols j1 to j2.
 * Modified by Peter Adams, August 1992.
 * The original version is at the bottom of this file, commented out.
 */
{
	char **str, *buff;
	unsigned int tmp, i, j, nrow, ncol, len, nstr = 0;
	int *colwidth;
	unsigned int ct;
	
	if(!Mptr) return;
	nrow = i2 - i1 + 1;
	ncol = j2 - j1 + 1;
	if (MAX((int)nrow, (int)ncol) > 50)
	{
		printf("(The number of rows or columns to be printed exceeds 50;\n");
		printf("there is no point in printing this matrix on the screen.)\n");
		return;
	}
	str = (char **)mmalloc(nrow * ncol * sizeof(char *));
	colwidth=(int *)mmalloc(ncol*sizeof(int));

	for(i=0; i<ncol; i++)
	  colwidth[i]=0;
	for (i = i1; i <= i2; i++)
	{
		for (j = j1; j <= j2; j++)
		{
			tmp = 1 + LENGTHI(elt(Mptr, i, j));
			buff = (char *)mmalloc(tmp * sizeof(char));
			SPRINTI(buff, elt(Mptr, i, j));
		    	if ((len = strlen(buff)) > colwidth[j-j1])
				colwidth[j-j1] = len;
			str[nstr++] = strcpy((char *)mmalloc(len + 1), buff);
			ffree(buff, tmp * sizeof(char));
		}
	}
	ct=0;
	for (i = i1; i <= i2; i++)
	{
		for (j = j1; j <= j2; j++)
		{
	    		printf("%*s ", colwidth[j-j1], str[ct]);
	    		ffree(str[ct], (unsigned int) (strlen(str[ct]) + 1) * sizeof(char));
	    		if ((ct % ncol) == (ncol - 1))
				printf("\n");
	    		ct++;
		}
	}
	ffree((char *) str, nrow * ncol * sizeof(char *));
	ffree((char *) colwidth, ncol*sizeof(int));
	return;
}

void FPRINTMATI(FILE *outfile, USI i1, USI i2, USI j1, USI j2, MPMATI *Mptr)
{
	char **str, *buff;
	unsigned int i, j, nrow, ncol, len, nstr = 0, tmp;
	int maxlen = 0;
	
	nrow = i2 - i1 + 1;
	ncol = j2 - j1 + 1;
	str = (char **)mmalloc(nrow * ncol * sizeof(char *));
	for (i = i1; i <= i2; i++)
	{
		for (j = j1; j <= j2; j++)
		{
			tmp = (1 + LENGTHI(elt(Mptr, i, j)));
			buff = (char *)mmalloc(tmp * sizeof(char));
			SPRINTI(buff, elt(Mptr, i, j));
		    	if ((len = strlen(buff)) > maxlen)
				maxlen = len;
			str[nstr++] = strcpy((char *)mmalloc((unsigned int)(len + 1)), buff);
			ffree(buff, tmp * sizeof(char));
		}
	}
	if (outfile != stdout)
		fprintf(outfile, "%u %u\n", nrow, ncol);
	for (i = 0; i < nstr; i++)
	{
	    fprintf(outfile, "%*s ", maxlen, str[i]);
	    ffree(str[i], (unsigned)(strlen(str[i])+1) * sizeof(char));
	    if ((i % ncol) == (ncol - 1))
		fprintf(outfile,"\n");
	}
	ffree((char *)str, nrow * ncol * sizeof(char *));
	return;
}

MPMATI *FINPUTMATFILEI_I()
{
	char filename[50];
	MPMATI *Mptr;
	FILE *infile;

	fprintf(stderr,"enter the name of your matrix file:");
	gets(filename);
	printf("filename : %s\n", filename);
	infile = Fopen_r(filename);
	if (infile == NULL)
	{
		printf("The file `%s' does not exist!\n",filename);
		GetReturn();
		Mptr = NULL;
	}
	else
	{
		Mptr = FINPUTMATI(infile);
		fclose(infile);
		fprintf(stderr,"matrices from file %s successfully inputted\n", filename);
		GetReturn();
	}
	return (Mptr);
}

MPI *NEAREST_INTI(MPI *Aptr, MPI *Bptr)
/*
 * Y is the nearest integer to *Aptr/(*Bptr).
 */
{
	MPI *X, *Y, *Z, *Tmp1, *Tmp2;

	Y = INT(Aptr, Bptr);
	X = MULTI(Bptr, Y);
	Z = SUBI(Aptr, X);
	FREEMPI(X);
	Tmp1 = Z;
	Z = MULT_I(Z, 2);
	FREEMPI(Tmp1);
	if (RSV(Z, Bptr) == 1)
	{
		Tmp2 = ONEI();
		Tmp1 = Y;
		Y = ADDI(Y, Tmp2);
		FREEMPI(Tmp2);
		FREEMPI(Tmp1);
	}
	FREEMPI(Z);
	return (Y);
}

/*
ORIGINAL VERSION OF PRINTMATI:
void PRINTMATI(i1, i2, j1, j2, Mptr)
MPMATI *Mptr;
unsigned int i1, i2, j1, j2;
{
	char **str, *buff;
	unsigned int tmp, i, j, nrow, ncol, len, maxlen = 0, nstr = 0;
	
	nrow = i2 - i1 + 1;
	ncol = j2 - j1 + 1;
	str = (char **)mmalloc(nrow * ncol * sizeof(char *));
	for (i = i1; i <= i2; i++)
	{
		for (j = j1; j <= j2; j++)
		{
			tmp = 1 + LENGTHI(elt(Mptr, i, j));
			buff = mmalloc(tmp * sizeof(char));
			SPRINTI(buff, elt(Mptr, i, j));
		    	if ((len = strlen(buff)) > maxlen)
				maxlen = len;
			str[nstr++] = strcpy(mmalloc(len + 1), buff);
			ffree(buff, tmp * sizeof(char));
		}
	}
	for (i = 0; i < nstr; i++)
	{
	    printf("%*s ", maxlen, str[i]);
	    ffree(str[i], (unsigned int) (strlen(str[i]) + 1) * sizeof(char));
	    if ((i % ncol) == (ncol - 1))
		printf("\n");
	}
	ffree((char *) str, nrow * ncol * sizeof(char *));
	return;
}
*/

MPI *GCD_ARRAY(MPI *M[], USI n)
/*
 * *Aptr = GCD(M[0], ..., M[n - 1]).
 */
{
	MPI *L, *Aptr;
	unsigned int i;

	Aptr = COPYI(M[0]);
	if (n == 1)
		return Aptr;
	else
	{
		for ( i = 1; i <= n - 1; i++)
		{
			L = Aptr;
			Aptr = GCD(L, M[i]);
			FREEMPI(L);
		}
		return Aptr;
	}
}
