/* pmm.c */

#include <stdio.h>
#include "integer.h"
#include "fun.h"

void PINSERTPM(int DEG, MPI *COEF, POLYI *Pptr, USI OPT, MPI *Mptr)
/*
 * inserts a copy of a single term *Qptr into polynomial *Pptr.
 * It has two options that control the insertion in case a term of *Pptr already
 * has the degree of *Qptr, If OPT = 0, then the term is replaced; if OPT = 1,
 * then the coefficient of *Qptr is added to that of the corresponding term of
 *  *Pptr (mod *Mptr).
 */
{
	POLYI CURSOR, Ttmp;
	MPI *TempI;

	if (*Pptr == NULL)
	{
		*Pptr = CREATEI();
		(*Pptr)->DEG = DEG;
		(*Pptr)->COEF = COPYI(COEF);
		(*Pptr)->NEXT = NULL;
		return;
	}
	else
	{
		CURSOR = *Pptr;
		while ((CURSOR->DEG > DEG) && (CURSOR->NEXT != NULL))
			CURSOR = CURSOR->NEXT;
		if (CURSOR->DEG < DEG)
		{
			/* Qptr inserted before CURSOR */
			Ttmp = CREATEI();
			Ttmp->DEG = CURSOR->DEG;
			Ttmp->COEF = CURSOR->COEF;
			Ttmp->NEXT = CURSOR->NEXT;
			CURSOR->DEG = DEG;
			CURSOR->COEF = COPYI(COEF);
			CURSOR->NEXT = Ttmp;
			return;
		}
		else if (CURSOR->DEG == DEG)
		{
			if (OPT == 0)
				CURSOR->COEF = COPYI(COEF);
			else
			{
				TempI = CURSOR->COEF;
				CURSOR->COEF = ADDM(CURSOR->COEF, COEF, Mptr);
				FREEMPI(TempI);
			}
			return;
		}
		else
		{
			/* CURSOR->DEG > Qptr->DEG and CURSOR->NEXT === NULL */
			/* Qptr inserted at end */
			Ttmp = CREATEI();
			Ttmp->DEG = DEG;
			Ttmp->COEF = COPYI(COEF);
			Ttmp->NEXT = NULL;
			CURSOR->NEXT = Ttmp;
			return;
		}
	}
}
 
POLYI SCALARPM(MPI *Cptr, POLYI Pptr, MPI *Mptr)
/*
 * multiplying the polynomial Pptr by the residue *Cptr, 0 <= *Cptr < *Mptr.
 * for use with polynomials mod *Mptr, 1 < *Mptr.
 */
{
	POLYI Sptr, Rtmp;
	MPI *TempI;

	if (Pptr == NULL || Cptr->S == 0)
		Sptr = NULL;
	else
	{
		Sptr = COPYPI(Pptr);
		Rtmp = Sptr;
		while (Rtmp != NULL)
		{
			TempI = Rtmp->COEF;
			Rtmp->COEF = MULTM(Rtmp->COEF, Cptr, Mptr);
			FREEMPI(TempI);
			Rtmp = Rtmp->NEXT;
		}
	}
	return (Sptr);
}

POLYI ADDPM(POLYI Pptr, POLYI Qptr, MPI *Mptr)
/*
 * returns the sum of two polynomials.
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 */
{
	POLYI CURSOR, Sptr;

	Sptr = COPYPI(Pptr);
	CURSOR = Qptr;
	while (CURSOR != NULL)
	{
		PINSERTPM(CURSOR->DEG, CURSOR->COEF, &Sptr, 1, Mptr);
		CURSOR = CURSOR->NEXT;
	}
	PURGEPI(&Sptr);
	return (Sptr);
}

POLYI SUBPM(POLYI Pptr, POLYI Qptr, MPI *Mptr)
/*
 * returns the difference Pptr - Qptr of two polynomials.
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 */
{
	POLYI CURSOR, Sptr;
	MPI *TempI;

	Sptr = COPYPI(Pptr);
	if (Qptr == NULL)
		return (Sptr);
	CURSOR = Qptr;
	while (CURSOR != NULL)
	{
		TempI = MINUSM(CURSOR->COEF, Mptr);
		PINSERTPM(CURSOR->DEG, TempI, &Sptr, 1, Mptr);
		FREEMPI(TempI);
		CURSOR = CURSOR->NEXT;
	}
	PURGEPI(&Sptr);
	return (Sptr);
}

POLYI MULTPM(POLYI Pptr, POLYI Qptr, MPI *Mptr)
/*
 * returns the product of two polynomials.
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 */
{
	int k;
	POLYI Sptr, CURSOR_P, CURSOR_Q;
	MPI *TempI;

	if (Pptr == NULL || Qptr == NULL)
		return (NULL);
	Sptr = NULL;
	CURSOR_P = Pptr;
	while (CURSOR_P != NULL)
	{
		CURSOR_Q = Qptr;
		while (CURSOR_Q != NULL)
		{
			k = CURSOR_P->DEG + CURSOR_Q->DEG;
			TempI = MULTM(CURSOR_P->COEF, CURSOR_Q->COEF, Mptr);
			PINSERTPM(k, TempI, &Sptr, 1, Mptr);
			FREEMPI(TempI);
			CURSOR_Q = CURSOR_Q->NEXT;
		}
		CURSOR_P = CURSOR_P->NEXT;
	}
	PURGEPI(&Sptr);
	return (Sptr);
}

MPI *VALPM(POLYI Pptr, MPI *Cptr, MPI *Mptr)
/*
 * Evaluating the polynomial Pptr at the residue *Cptr, 0 <= *Cptr < *Mptr.
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 */
{
	int k;
	MPI *B, *Aptr, *TempI;
	POLYI CURSOR;

	if (Pptr == NULL)
		return ZEROI();
	else
		Aptr = COPYI(Pptr->COEF);
	CURSOR = Pptr;
	while (CURSOR != NULL)
	{
		if (CURSOR->NEXT == NULL)
			k = DEGREEPI(CURSOR);
		else
			k = DEGREEPI(CURSOR) - DEGREEPI(CURSOR->NEXT);
		B = MPOWER_M(Cptr, (unsigned int)k, Mptr);
		TempI = Aptr;
		Aptr = MULTM(Aptr, B, Mptr);
		FREEMPI(B);
		FREEMPI(TempI);
		CURSOR = CURSOR->NEXT;
		if (CURSOR != NULL)
		{
			TempI = Aptr;
			Aptr = ADDM(Aptr, CURSOR->COEF, Mptr);
			FREEMPI(TempI);
		}
	}
	return (Aptr);
}

POLYI DIVPM(POLYI Fptr, POLYI Gptr, MPI *Mptr)
/*
 * division of polynomials: see Knuth, Vol 2, p.402
 * and H. Flanders, Scientific Pascal, p.510.
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 */
{
	POLYI Htmp, CURSOR, Qptr, Temp, F;
	int j, d;
	MPI *A, *G, *TempI;

	Qptr = NULL;
	d = DEGREEPI(Gptr);
	j = DEGREEPI(Fptr) - d;
	if (j < 0)
		return (Qptr);
	else
	{
		F = COPYPI(Fptr);
		G =LEADPI(Gptr);
		while ((j = DEGREEPI(F)) >= d)
		{
			j = j - d;
			A = LEADPI(F);
			TempI = A;
			A = DIVM(A, G, Mptr);
			FREEMPI(TempI);
			Htmp = COPYPI(Gptr);
			CURSOR = Htmp;
			do
			{
				CURSOR->DEG = j + CURSOR->DEG;
				TempI = CURSOR->COEF;
				CURSOR->COEF = MINUSM(CURSOR->COEF, Mptr);
				FREEMPI(TempI);
				TempI = CURSOR->COEF;
				CURSOR->COEF = MULTM(CURSOR->COEF, A, Mptr);
				FREEMPI(TempI);
				CURSOR = CURSOR->NEXT;
			}
			while (CURSOR != NULL);
			Temp = ADDPM(F, Htmp, Mptr);
			DELETE_PI(Htmp);
			DELETE_PI(F);
			F = Temp;
			PINSERTPM(j, A, &Qptr, 0, Mptr);
			FREEMPI(A);
		}
		DELETE_PI(F);
		FREEMPI(G);
		return (Qptr);
	}
}

POLYI MODPM(POLYI Fptr, POLYI Gptr, MPI *Mptr)
/*
 * division of polynomials: see Knuth, Vol 2, p.402
 * and H. Flanders, Scientific Pascal, p.510.
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 */
{
	POLYI Htmp, CURSOR, Temp, F;
	int j, d;
	MPI *A, *B, *G, *Temp1I;

	d = DEGREEPI(Gptr);
	j = DEGREEPI(Fptr) - d;
	if (j < 0)
		return COPYPI(Fptr);
	else if (d == 0)
		return NULL;
	else if (d == 1)
	{
		if (Gptr->NEXT == NULL)
		{
			Temp1I = ZEROI();
			B = VALPM(Fptr, Temp1I, Mptr);
			FREEMPI(Temp1I);
			F = CONSTANTPM(B);
			FREEMPI(B);
			return (F);
		}
		else
		{
			B = DIVM((Gptr->NEXT)->COEF, Gptr->COEF, Mptr);
			Temp1I = B;
			B = MINUSM(B, Mptr);
			FREEMPI(Temp1I);
			Temp1I = B;
			B = VALPM(Fptr, B, Mptr);
			FREEMPI(Temp1I);
			F = CONSTANTPM(B);
			FREEMPI(B);
			return (F);
		}
	}
	else
	{
		F = COPYPI(Fptr);
		G = LEADPI(Gptr);
		while ((j = DEGREEPI(F)) >= d)
		{
			j = j - d;
			A = LEADPI(F);
			Temp1I = A;
			A = DIVM(A, G, Mptr);
			FREEMPI(Temp1I);
			Htmp = COPYPI(Gptr);
			CURSOR = Htmp;
			do
			{
				CURSOR->DEG = j + CURSOR->DEG;
				Temp1I = CURSOR->COEF;
				CURSOR->COEF = MINUSM(CURSOR->COEF, Mptr);
				FREEMPI(Temp1I);
				Temp1I = CURSOR->COEF;
				CURSOR->COEF = MULTM(CURSOR->COEF, A, Mptr);
				FREEMPI(Temp1I);
				CURSOR = CURSOR->NEXT;
			}
			while (CURSOR != NULL);
			Temp = F;
			F = ADDPM(Temp, Htmp, Mptr);
			DELETE_PI(Temp);
			DELETE_PI(Htmp);
			FREEMPI(A);
		}
		FREEMPI(G);
		return (F);
	}
}

POLYI MONICPM(POLYI Pptr, MPI *Mptr)
/*
 * converts Pptr to a monic polynomial by multiplying it by 1/LEADPM(Pptr).
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 */
{
	MPI *G, *TempI;
	POLYI P;

	G = LEADPI(Pptr);
	TempI = G;
	G = INVERSEM(G, Mptr);
	FREEMPI(TempI);
	P = SCALARPM(G, Pptr, Mptr);
	FREEMPI(G);
	return (P);
}

POLYI GCDPM(POLYI Pptr, POLYI Qptr, MPI *Mptr)
/*
 * returns the gcd of polynomials Pptr and Qptr.
 * see Knuth, Vol 2, p.403-408 and H. Flanders, Scientific Pascal, p.510.
 * for use with polynomials mod Mptr, 1 < Mptr. 
 */
{
	POLYI Ptmp, Qtmp, Rtmp;
	int j;
	
	if (Qptr == NULL)
	{
		if (Pptr == NULL)
			return (NULL);
		else
			return (MONICPM(Pptr, Mptr));
	}
	else if (Pptr == NULL)
	{
		if (Qptr == NULL)
			return (NULL);
		else
			return (MONICPM(Qptr, Mptr));
	}
	else
	{
		Ptmp = COPYPI(Pptr);
		Qtmp = COPYPI(Qptr);
		j = DEGREEPI(Qtmp); /* to keep the compiler quiet */
		while ((Qtmp != NULL) && ((j = DEGREEPI(Qtmp)) != 0))
		{
			Rtmp =	MODPM(Ptmp, Qtmp, Mptr);
			DELETE_PI(Ptmp);
			Ptmp = Qtmp;
			Qtmp = Rtmp;
		}
		if (j == 0)
		{
			Qtmp->COEF = ONEI();
			DELETE_PI(Ptmp);
			return (Qtmp);
		}
		else
		{
			Qtmp = MONICPM(Ptmp, Mptr);
			DELETE_PI(Ptmp);
			return (Qtmp);
		}
	}
}

POLYI DERIVPM(POLYI Pptr, MPI *Mptr)
/*
 * returns the derivative of a polynomial.
 * for use with polynomials mod Mptr, 1 < Mptr.
 */
{
	POLYI CURSOR, Sptr;
	MPI *A, *TempI;

	if (DEGREEPI(Pptr) == 0)
		return (NULL);
	else
	{
		Sptr = COPYPI(Pptr);
  		CURSOR = Sptr;
		while (CURSOR != NULL)
		{
			A = CHANGE((unsigned int)CURSOR->DEG);
			TempI = CURSOR->COEF;
			CURSOR->COEF = MULTM(CURSOR->COEF, A, Mptr);
			FREEMPI(TempI);
			FREEMPI(A);
			CURSOR->DEG = CURSOR->DEG - 1;

			if (CURSOR->NEXT == NULL)
				break;
			else if ((CURSOR->NEXT)->DEG == 0)
			{
				DELETE_PI(CURSOR->NEXT);
				CURSOR->NEXT = NULL;
				break;
			}
			else
				CURSOR = CURSOR->NEXT;
		}
		PURGEPI(&Sptr);
		return (Sptr);
	}
}

POLYI POWERPM(POLYI Pptr, USI n, MPI *Mptr)
/*
 * returns Pptr ^ n mod *Mptr, where 0 <= n < R0 * R0.
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 */
{
	POLYI B, Eptr, Tmp;

	Eptr = ONEPI();
	if (n == 0)
		return (Eptr);
	B = COPYPI(Pptr);
	while (1)
        {
		if (n & 1) 
		{
			Tmp = Eptr;
			Eptr = MULTPM(Tmp, B, Mptr);
			DELETE_PI(Tmp);
			if (n == 1)
			{
				DELETE_PI(B);
				return (Eptr);
			}
		}
		Tmp = B;
		B = MULTPM(Tmp, Tmp, Mptr);
		DELETE_PI(Tmp);
		n = n >> 1;
	}
}

POLYI CONSTANTPM(MPI *Aptr)
/*
 * returns the constant polynomial whose constant term is the residue *Aptr.
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 * 0 <= *Aptr < *Mptr.
 */
{
	POLYI Pptr;	

	if (Aptr->S == 0)
		return (NULL);
	else
	{
		Pptr = CREATEI();
		Pptr->DEG = 0;
		Pptr->COEF = COPYI(Aptr);
		Pptr->NEXT = NULL;
		return (Pptr);
	}
}

POLYI LINEARPM(MPI *Aptr, MPI *Mptr)
/*
 * 0 <= *Aptr < *Mptr. 
 * returns the polynomial x + *Aptr.
 */
{
	POLYI Pptr;
	MPI *TempI;

	Pptr = CONSTANTPM(Aptr);
	TempI = ONEI();
	PINSERTPM(1, TempI, &Pptr, 0, Mptr);
	FREEMPI(TempI);
	return (Pptr);
}

POLYI POWERPPM(POLYI Pptr, MPI *Bptr, POLYI Qptr, MPI *Mptr)
/*
 *  returns Pptr ^ *Bptr, mod Qptr, where 0 <= *Bptr.
 * for use with polynomials mod *Mptr, 1 < *Mptr. 
 */
{
	POLYI B, Eptr, Tmp;
	MPI *N, *TempI;

	Eptr = ONEPI();
	if (Bptr->S == 0)
		return (Eptr);
	N = COPYI(Bptr);
	B = COPYPI(Pptr);
	while (1)
        {
		if (N->V[0] & 1) 
		{
			Tmp = Eptr;
			Eptr = MULTPM(Tmp, B, Mptr);
			DELETE_PI(Tmp);
			Tmp = Eptr;
			Eptr = MODPM(Tmp, Qptr, Mptr);
			DELETE_PI(Tmp);
			if (N->S == 1 && N->D == 0 && N->V[0] == 1)
			{
				FREEMPI(N);
				DELETE_PI(B);
				return (Eptr);
			}
		}
		Tmp = B;
		B = MULTPM(Tmp, Tmp, Mptr);
		DELETE_PI(Tmp);
		Tmp = B;
		B = MODPM(Tmp, Qptr, Mptr);
		DELETE_PI(Tmp);
		TempI = N;
		N = INT0_(N, 2);
		FREEMPI(TempI);
	}
}

POLYI EUCLIDPM(POLYI Pptr, POLYI Qptr, POLYI *Hptr, POLYI *Kptr, MPI *Mptr)
/*
 * input: polynomials mod prime *Mptr, 
 * gcd(Pptr, Qptr) = Hptr * Pptr + Kptr * Qptr.
 * all mod *Mptr. 
 */
{
	POLYI Q, A, B, C, H1, H2, K1, K2, L1, L2, Tmp, Tmp1, Tmp2;
	POLYI Temp1, Temp2;
	MPI *G, *TempI;

	if (Qptr == NULL)
	{
		if (Pptr != NULL)
		{
			G = LEADPI(Pptr);
			TempI = G;
			G = INVERSEM(G, Mptr);
			FREEMPI(TempI);
			*Hptr = CONSTANTPM(G);
			FREEMPI(G);
			*Kptr = NULL;
			return (MONICPM(Pptr, Mptr));
		}
		else
		{
			*Hptr = NULL;
			*Kptr = NULL;
			return (NULL);
		}
	}
	A = COPYPI(Pptr);
	B = COPYPI(Qptr);
	C = MODPM(A, B, Mptr);
	if (C == NULL)
	{
		*Hptr = NULL;
		G = LEADPI(Qptr);
		TempI = G;
		G = INVERSEM(G, Mptr);
		FREEMPI(TempI);
		*Kptr = CONSTANTPM(G);
		FREEMPI(G);
		DELETE_PI(A);
		DELETE_PI(B);
		return (MONICPM(Qptr, Mptr));
	}
	TempI = ONEI();
	L1 = CONSTANTPM(TempI);
	K1 = NULL;
	L2 = NULL;
	K2 = CONSTANTPM(TempI);
	FREEMPI(TempI);
	while (C != NULL)
	{
		Q = DIVPM(A, B, Mptr);
		DELETE_PI(A);
		A = B;
		B = C;
		C = MODPM(A, B, Mptr);
		Tmp1 = MULTPM(Q, K1, Mptr);
		Tmp2 = MULTPM(Q, K2, Mptr);
		DELETE_PI(Q);
		H1 = SUBPM(L1, Tmp1, Mptr);
		H2 = SUBPM(L2, Tmp2, Mptr);
		DELETE_PI(Tmp1);
		DELETE_PI(Tmp2);
		DELETE_PI(L1);
		L1 = K1;
		DELETE_PI(L2);
		L2 = K2;
		K1 = H1;
		K2 = H2;
	}
	G = LEADPI(B);
	TempI = G;
	G = INVERSEM(G, Mptr);
	FREEMPI(TempI);
	Temp1 = SCALARPM(G, K1, Mptr);
	Temp2 = SCALARPM(G, K2, Mptr);
	FREEMPI(G);
	*Hptr = MODPM(Temp1, Qptr, Mptr);
	*Kptr = MODPM(Temp2, Pptr, Mptr);
	DELETE_PI(Temp1);
	DELETE_PI(Temp2);
	DELETE_PI(K1);
	DELETE_PI(K2);
	DELETE_PI(L1);
	DELETE_PI(L2);
	DELETE_PI(A);
	DELETE_PI(C);
	Tmp = MONICPM(B, Mptr);
	DELETE_PI(B);
	return (Tmp);
}

POLYI CHANGETOPM(POLYI Pptr, MPI *Mptr)
/*
 * converts a polynomial with MPI coefficients into one with coefficients mod 
 * *Mptr.
 */
{
	POLYI CURSOR, P;
	MPI *TempI;

	P = NULL;
	CURSOR = Pptr;
	while (CURSOR != NULL)
	{
		TempI = MOD(CURSOR->COEF, Mptr);
		PINSERTPM(CURSOR->DEG, TempI, &P, 0, Mptr);
		FREEMPI(TempI);
		CURSOR = CURSOR->NEXT;
	}	
	PURGEPI(&P);
	return (P);
}

MPI *HENSELPM(POLYI Fptr, POLYI u1ptr, POLYI u2ptr, POLYI v1ptr, POLYI v2ptr, POLYI *U1ptr, POLYI *U2ptr, POLYI *V1ptr, POLYI *V2ptr, MPI *Qptr)
/*
 * input: polynomials mod *Qptr : u1ptr, u2ptr, v1ptr, v2ptr, with Fptr a 
 * polynomial with MPI coefficients; gcd(*Qptr, LEADPI(Fptr)) = 1;
 * LEADPI(Fptr) = LEADPm(u1ptr) mod *Qptr; u2ptr monic;
 * DEGREEPI(viptr) < DEGREEPI(uiptr) for i = 1, 2;
 * u1ptr*v2ptr + u2ptr*v1ptr = 1 mod *Qptr;
 * output: polynomials mod *Qptr^2 : U1ptr, U2ptr, V1ptr, V2ptr, with
 * LEADPI(Fptr) = LEADPI(U1ptr) mod *Qptr^2; U2ptr monic;
 * U1ptr*V2ptr + U2ptr*V1ptr = 1 mod *Qptr^2;
 * DEGREEPI(Viptr) < DEGREEPI(Uiptr) for i = 1, 2;
 * Uiptr = uiptr mod *Qptr, Viptr = viptr mod *Qptr, i = 1, 2;
 * *Mptr = *Qptr^2. 
 * The algorithm is the quadratic lift version of `Hensel lifting' and is 
 * due to Zassenhaus. See E. Kaltofen, Factorization of polynomials, pp. 95-113,
 * Computing Supplement 4, Springer 1982.
 */
{
	POLYI w1, w2, f, t, CURSOR, one, b, Temp, Temp1, Temp2, Temp3;
	MPI *Mptr, *TempI;

	Mptr = MULTI(Qptr, Qptr);
	/* Step H1 */
	Temp = COPYPI(u1ptr);
	TempI = Temp->COEF;
	Temp->COEF = MOD(Fptr->COEF, Mptr);
	FREEMPI(TempI);
	Temp1 = MULTPM(Temp, u2ptr, Mptr);
	f = CHANGETOPM(Fptr, Mptr);
	t = SUBPM(f, Temp1, Mptr);
	/* DEGREEPI(t) < DEGREEPI(Fptr) */
	DELETE_PI(Temp1);
	DELETE_PI(f);
	CURSOR = t;
	/* the coefficients of t are divisible by q */
	while (CURSOR != NULL)
	{
		TempI = CURSOR->COEF;
		CURSOR->COEF = INT0(CURSOR->COEF, Qptr); 
		FREEMPI(TempI);
		CURSOR = CURSOR->NEXT;
	}
	Temp1 = MULTPM(t, v1ptr, Qptr);
	Temp2 = MULTPM(t, v2ptr, Qptr);
	w1 = MODPM(Temp1, u1ptr, Qptr);
	w2 = MODPM(Temp2, u2ptr, Qptr);
	DELETE_PI(Temp1);
	DELETE_PI(Temp2);
	Temp1 = SCALARPM(Qptr, w1, Mptr);
	Temp2 = SCALARPM(Qptr, w2, Mptr);
	DELETE_PI(w1);
	DELETE_PI(w2);
	*U1ptr = ADDPM(Temp, Temp1, Mptr);
	*U2ptr = ADDPM(u2ptr, Temp2, Mptr);
	DELETE_PI(Temp1);
	DELETE_PI(Temp2);
	/* at this stage t = w1*v2ptr + w2*v1ptr mod *Qptr */
	DELETE_PI(t);

	/* Step H2 */
	Temp1 = MULTPM(*U1ptr, v2ptr, Mptr);
	Temp2 = MULTPM(*U2ptr, v1ptr, Mptr);
	Temp3 = ADDPM(Temp1, Temp2, Mptr);
	DELETE_PI(Temp1);
	DELETE_PI(Temp2);
	TempI = ONEI();
	one = CONSTANTPM(TempI);
	FREEMPI(TempI);
	b = SUBPM(one, Temp3, Mptr);
	/* DEGREEPI(b) < DEGREEPI(Fptr) */
	DELETE_PI(one);
	DELETE_PI(Temp3);
	CURSOR = b;
	/* the coefficients of b are divisible by *Qptr */
	while (CURSOR != NULL)
	{
		TempI = CURSOR->COEF;
		CURSOR->COEF = INT0(CURSOR->COEF, Qptr); 
		FREEMPI(TempI);
		CURSOR = CURSOR->NEXT;
	}
	Temp1 = MULTPM(b, v1ptr, Qptr);
	Temp2 = MULTPM(b, v2ptr, Qptr);
	DELETE_PI(b);
	w1 = MODPM(Temp1, u1ptr, Qptr);
	w2 = MODPM(Temp2, u2ptr, Qptr);
	DELETE_PI(Temp);
	DELETE_PI(Temp1);
	DELETE_PI(Temp2);
	Temp1 = SCALARPM(Qptr, w1, Mptr);
	Temp2 = SCALARPM(Qptr, w2, Mptr);
	DELETE_PI(w1);
	DELETE_PI(w2);
	*V1ptr = ADDPM(v1ptr, Temp1, Mptr);
	*V2ptr = ADDPM(v2ptr, Temp2, Mptr);
	DELETE_PI(Temp1);
	DELETE_PI(Temp2);
	/* at this stage 1 = w1*v2ptr + w2*v1ptr mod *Mptr */
	return (Mptr);

}

MPI *HENSEL1PM(POLYI Fptr, POLYI u1ptr, POLYI u2ptr, POLYI v1ptr, POLYI v2ptr, POLYI *U1ptr, POLYI *U2ptr, MPI *Qptr, USI p)
/*
 * input: polynomials mod *Qptr : u1ptr, u2ptr, v1ptr, v2ptr, with Fptr a 
 * polynomial with MPI coefficients; gcd(*Qptr, LEADPI(Fptr)) = 1;
 * LEADPI(Fptr) = LEADPm(u1ptr) mod *Qptr; u2ptr monic;
 * DEGREEPI(viptr) < DEGREEPI(uiptr) for i = 1, 2;
 * u1ptr*v2ptr + u2ptr*v1ptr = 1 mod p;
 * output: polynomials mod *Qptr^2 : U1ptr, U2ptr, with
 * LEADPI(Fptr) = LEADPI(U1ptr) mod *Mptr; U2ptr monic;
 * U1ptr*v2ptr + U2ptr*v1ptr = 1 mod p;
 * Uiptr = uiptr mod p, i = 1, 2;
 * *Mptr = *Qptr * p. 
 * The algorithm is the linear lift version of `Hensel lifting'.
 * See E. Kaltofen, Factorization of polynomials, pp. 100-101,
 * Computing Supplement 4, Springer 1982.
 */
{
	POLYI w1, w2, f, t, CURSOR, Temp, Temp1, Temp2;
	MPI *Mptr, *TempI;

	Mptr = MULT_I(Qptr, (int)p);
	/* Step H1 */
	Temp = COPYPI(u1ptr);
	Temp->COEF = MOD(Fptr->COEF, Mptr);
	Temp1 = MULTPM(Temp, u2ptr, Mptr);
	f = CHANGETOPM(Fptr, Mptr);
	t = SUBPM(f, Temp1, Mptr);
	/* DEGREEPI(t) < DEGREEPI(Fptr) */
	DELETE_PI(Temp1);
	DELETE_PI(f);
	CURSOR = t;
	/* the coefficients of t are divisible by q */
	while (CURSOR != NULL)
	{
		TempI = CURSOR->COEF;
		CURSOR->COEF = INT0(CURSOR->COEF, Qptr); 
		FREEMPI(TempI);
		CURSOR = CURSOR->NEXT;
	}
	Temp1 = MULTPM(t, v1ptr, Qptr);
	Temp2 = MULTPM(t, v2ptr, Qptr);
	w1 = MODPM(Temp1, u1ptr, Qptr);
	w2 = MODPM(Temp2, u2ptr, Qptr);
	DELETE_PI(Temp1);
	DELETE_PI(Temp2);
	Temp1 = SCALARPM(Qptr, w1, Mptr);
	Temp2 = SCALARPM(Qptr, w2, Mptr);
	DELETE_PI(w1);
	DELETE_PI(w2);
	*U1ptr = ADDPM(Temp, Temp1, Mptr);
	*U2ptr = ADDPM(u2ptr, Temp2, Mptr);
	DELETE_PI(Temp1);
	DELETE_PI(Temp2);
	/* at this stage t = w1*v2ptr + w2*v1ptr mod *Qptr */
	DELETE_PI(t);
	DELETE_PI(Temp);
	return (Mptr);
}
